<?php

defined('MOODLE_INTERNAL') || die;

require_once("$CFG->dirroot" . "/course/externallib.php");
require_once("$CFG->dirroot" . "/course/lib.php");
require_once("$CFG->dirroot" . "/course/externallib.php");
require_once("$CFG->dirroot" . "/enrol/externallib.php");
require_once("$CFG->dirroot" . "/grade/lib.php");
require_once("$CFG->dirroot" . "/grade/querylib.php");
require_once("$CFG->dirroot" . "/enrol/manual/externallib.php");
require_once("$CFG->dirroot" . "/lib/moodlelib.php");
require_once("$CFG->dirroot" . "/local/stag/attributes.php");

/**
 * STAG-related course functions
 *
 * @package    local
 * @category   external
 */

class local_stag_course_external extends external_api {

    const CATEGORY_TEMPLATE_NAME = "Course Template Category";

    private static $stag_attributes = array(
        'department'    => array(
            'type'  => PARAM_TEXT,
            'desc'  => "identification of the department in STAG"),
        'subject'       => array(
            'type'  => PARAM_TEXT,
            'desc'  => "Subject name taken from STAG"),
        'year'          => array(
            'type'  => PARAM_INT,
            'desc'  => "Year of the course"),
        'roakidno'      => array(
            'type'  => PARAM_ALPHANUMEXT,
            'desc'  => "ROAKIDNO value from STAG"),
        'sync'          => array(
            'type'  => PARAM_BOOL,
            'desc'  => "whether course is synchronized with STAG"),
        );

    private static $course_reset_flags = array(
        'reset_events',
        'reset_logs',
        'reset_notes',
        'reset_comments',
        'reset_completion',
        'delete_blog_associations',
        'reset_roles_overrides',
        'mform_showadvanced_last',
        'reset_roles_local',
        'reset_gradebook_items',
        'reset_groups_remove',
        'reset_groupings_remove',
    );

		/**
		 * @return \external_function_parameters
		 */
    public static function stag_create_course_parameters() {
        return new \external_function_parameters(array(
            'course' => new external_single_structure(array(
                'categoryid'  => new external_value(PARAM_INT, "Course category"),
                'shortname'   => new external_value(PARAM_TEXT, "Short name of course"),
                'full_name'   => new external_value(PARAM_TEXT, "Full name of course"),
                'annotation'  => new external_value(\PARAM_RAW, "Course annotation"),
                'start_date'  => new external_value(PARAM_ALPHANUMEXT, "Beginning of course"),

                /*
                 * unfortunately Moodle doesn't allow one to embed multiple single_structure, so
                 * we need to cheat a bit here and check that at most a single stag_id is received
                 */
                'stag_id'          => new external_multiple_structure(new external_single_structure(array(
                    'department' => new external_value(self::$stag_attributes['department']['type'], "Department"),
                    'subject'    => new external_value(self::$stag_attributes['subject']['type'], "Subject"),
                    'year'       => new external_value(PARAM_INT, "Year of variant"),
                    )), "STAG course identifier", VALUE_OPTIONAL),
                'stag_roakidno' => new external_value(PARAM_ALPHANUMEXT, "ROAKIDNO value from STAG", VALUE_OPTIONAL),
                'stag_sync'   => new external_value(PARAM_BOOL, "Whether to synchronize with STAG", VALUE_DEFAULT, True),
                'lookup_predecessor' => new external_value(PARAM_BOOL, "Whether to create course based on its predecessor", VALUE_DEFAULT, True),
                ))));
    }

		/**
		 *
		 * @global \moodle_database $DB
		 * @global \moodle_page $PAGE
		 * @param type $course_req
		 * @return integer course ID
		 * @throws \moodle_exception
		 * @throws \Exception
		 */
    public static function stag_create_course($course_req) {
        global $DB;

        $params = self::validate_parameters(self::stag_create_course_parameters(),
                       array('course' => $course_req));

        $course = $params['course'];

        $stag_id = null;
        if (isset($course['stag_id'])){
            $stag_id = $course['stag_id'][0];
				}

        $context = get_context_instance(CONTEXT_COURSECAT, $course['categoryid']);
        if (empty($context)){
            throw new \moodle_exception('incorrectcategory', 'local_stag', '', $course['categoryid']);
				}

        self::validate_context($context);

        require_capability('moodle/course:create', $context);
        require_capability('enrol/guest:config', $context); /* required to set the guest password */

        $template = self::find_course_template();
        if ($template == null){
            throw new \moodle_exception('missingtemplate', 'local_stag');
				}

        $last_run = null;
        if ($stag_id && $course['lookup_predecessor'] === True){
            $last_run = self::find_last_run($stag_id);
				}

        $master_id = ($last_run === null) ? $template->id : $last_run;

        $transaction = $DB->start_delegated_transaction();

        global $PAGE;
        $old_PAGE = $PAGE;
        $PAGE = new \moodle_page();

        $op = array(array('name'=>'blocks', 'value'=>1),
        array('name'=>'activities', 'value'=>1),
        array('name'=>'filters', 'value'=>1),
        array('name'=>'users', 'value'=>0),
        array('name'=>'role_assignments', 'value'=>0),
        array('name'=>'comments', 'value'=>0),
        array('name'=>'logs', 'value'=>0)
        );


        try {
            $course_info = \core_course_external::duplicate_course($master_id, $course['full_name'],
                    local_stag_course_external::get_unique_shortname($course['shortname']), $course['categoryid'],1,$op);
        } catch (\Exception $e) {
            $PAGE = $old_PAGE;
            throw $e;
        }
        $PAGE = $old_PAGE;

        $course_id = $course_info['id'];

        self::reset_course($course_id);

        $new_course = $DB->get_record("course", array('id' => $course_id));
        $new_course->summary = $course['annotation'];

        $start_date = date_parse($course['start_date']);
        if ($start_date['error_count'] > 0){
            throw new \moodle_exception('wrongdateformat', 'local_stag');
				}
        $new_course->startdate = make_timestamp($start_date['year'], $start_date['month'], $start_date['day']);

        $new_course->visible = $template->visible;

        /* see enrol/guest/lib.php */
        $instances = $DB->get_records('enrol', array('courseid'=>$course_id, 'enrol'=>'guest'));
        $pwd = generate_password();
        foreach ($instances as $instance) {
            $i = $instance->id;
            $new_course->{'enrol_guest_password_'.$i} = $pwd;
            $new_course->{'enrol_guest_status_'.$i} = $instance->status;
        }

        update_course($new_course);

        if ($stag_id) {
            \attribute_manager::set_attribute($course_id, self::$stag_attributes['department']['type'], 'department', $stag_id['department']);
            \attribute_manager::set_attribute($course_id, self::$stag_attributes['subject']['type'], 'subject', $stag_id['subject']);
            \attribute_manager::set_attribute($course_id, PARAM_INT, 'year', $stag_id['year']);

            $result = $DB->get_records_sql('SELECT t4.course_id, max(t4.value) as max_ver FROM {course_attributes} t1, {course_attributes} t2, {course_attributes} t3, {course_attributes} t4
                                            WHERE t1.course_id=t2.course_id AND t1.name = "department" AND t1.value=? AND t2.name = "subject" AND t2.value = ?
                                            AND t1.course_id=t3.course_id AND t3.name="year" AND t3.value=? AND t1.course_id=t4.course_id AND t4.name="ver" order by t4.value desc',
                                            array( $stag_id['department'] , $stag_id['subject'] , $stag_id['year'] ));

            $max_ver_key = array_keys($result);
            $ver = 1;
            if ($result[$max_ver_key[0]]->max_ver > 0) {
              $ver = $result[$max_ver_key[0]]->max_ver +1;
						}

            \attribute_manager::set_attribute($course_id, PARAM_INT, 'ver', $ver);
        }

        if (isset($course['stag_roakidno'])){
            \attribute_manager::set_attribute($course_id, PARAM_ALPHANUMEXT, 'roakidno', $course['stag_roakidno']);
				}

        \attribute_manager::set_attribute($course_id, PARAM_BOOL, 'sync', $course['stag_sync']);

        $transaction->allow_commit();

        return $course_id;
    }

		/**
		 * @return \external_value
		 */
    public static function stag_create_course_returns() {
        return new \external_value(PARAM_INT, "Course identifier");
    }

    /**
		 * @return \external_function_parameters
     */
    public static function stag_get_courses_parameters() {
        return \core_course_external::get_courses_parameters();
    }

		/**
		 * @global \stdClass $CFG
		 * @param array $options
		 * @return array
		 */
    public static function stag_get_courses($options = array()) {
        global $CFG;

        $courses = \core_course_external::get_courses($options);

        foreach ($courses as &$course) {
            foreach (self::$stag_attributes as $attr => $value) {
							$value=$value; // Vlahovic: jen abych vycistil hlaseni netbeans pri refaktoringu
                $val = attribute_manager::get_attribute($course['id'], $attr);
                if ($val) {
                    $attr_name = "stag_" . $attr;
                    $course[$attr_name] = $val;
                }
            }
            $course['url'] = "$CFG->wwwroot" . "/course/view.php?id=" .$course['id'];
        }

        return $courses;
    }

		/**
		 * @return \external_multiple_structure
		 */
    public static function stag_get_courses_returns() {
        $pars = \core_course_external::get_courses_returns();
        $p = &$pars->content->keys;

        foreach (self::$stag_attributes as $name => $val) {
            $attr_name = "stag_" . $name;
            $p[$attr_name] = new \external_value($val['type'], $val['desc'], VALUE_OPTIONAL);
        }
        $p['url'] = new \external_value(PARAM_URL, "Course URL");

        return $pars;
    }

    /**
		 * @return \external_function_parameters
     */
    public static function stag_enrol_users_parameters() {
        return new \external_function_parameters(array(
            'enrolment' => new \external_single_structure(array(
                'roleid' => new \external_value(PARAM_INT, 'Role to assign to the user'),
                'userid' => new \external_value(PARAM_INT, 'The user that is going to be enrolled'),
                'courseid' => new \external_value(PARAM_INT, 'The course to enrol the user role in'),
                'stag_attribute' => new \external_value(PARAM_ALPHANUMEXT, 'STAG atribute assigned to enrolment', VALUE_OPTIONAL)))));
    }

		/**
		 * @global \moodle_database $DB
		 * @param array $enrolment
		 * @return void
		 */
    public static function stag_enrol_users($enrolment) {
        global $DB;
        $params = self::validate_parameters(self::stag_enrol_users_parameters(),
                                array('enrolment' => $enrolment));

        $param = array(
            'roleid' => $params['enrolment']['roleid'],
            'userid' => $params['enrolment']['userid'],
            'courseid' => $params['enrolment']['courseid']);

        $transaction = $DB->start_delegated_transaction();

        \enrol_manual_external::enrol_users(array($param));

        $transaction->allow_commit();
    }

		/**
		 * @return null
		 */
    public static function stag_enrol_users_returns() {
        return null;
    }

    /**
     * Get course grade
		 * @return \external_function_parameters
     */
    public static function stag_get_grade_parameters() {
        return new \external_function_parameters(array(
            'grade_of' => new \external_single_structure(array(
                'username'    => new \external_value(PARAM_TEXT, "Username", VALUE_OPTIONAL),
                'course_id'   => new \external_value(PARAM_INT, "Course identifier"),
            ))));
    }

		/**
		 *
		 * @global \moodle_database $DB
		 * @param array $grade_of
		 * @return array
		 * @throws \moodle_exception
		 */
    public static function stag_get_grade($grade_of) {
        global $DB;

        $params = self::validate_parameters(self::stag_get_grade_parameters(),
                       array('grade_of' => $grade_of));

        $courseid = $params['grade_of']['course_id'];

        $context = get_context_instance(CONTEXT_SYSTEM);
        self::validate_context($context);
        require_capability('moodle/site:viewuseridentity', $context);

        $course_context = get_context_instance(CONTEXT_COURSE, $courseid);
        self::validate_context($course_context);
        require_capability('moodle/grade:viewall', $course_context);

        $users = array();
        $usernames = array();
        if (isset($params['grade_of']['username'])) {
            $username = $params['grade_of']['username'];
            $user = $DB->get_record('user', array('username'=>$username));
            if ($user == null){
                throw new \moodle_exception('nosuchuser', 'local_stag', '', $username);
						}
            $users[] = $user->id;
            $usernames[$user->id] = $username;
        } else {
            $all_users = \core_enrol_external::get_enrolled_users($courseid);
            foreach ($all_users as $user) {
                $users[] = $user['id'];
                $usernames[$user['id']] = $user['username'];
            }
        }

        $grades = grade_get_course_grades($courseid, $users);

        $return = array();
        foreach($grades->grades as $userid => $grade) {
            $return[] = array('username' => $usernames[$userid], 'grade' => $grade->str_grade);
        }

        return($return);
    }

		/**
		 * @return \external_multiple_structure
		 */
    public static function stag_get_grade_returns() {
        return new \external_multiple_structure(
            new \external_single_structure(array(
                'username' => new \external_value(PARAM_TEXT, "Username"),
                'grade' => new \external_value(PARAM_TEXT, "User's grade in course"),
                )));
    }

    /**
     * Get user identifier
		 * @return \external_function_parameters
     */
    public static function stag_get_user_id_parameters() {
        return new \external_function_parameters(array(
            'username' => new \external_value(PARAM_TEXT, "Username")));
    }

		/**
		 * @global \moodle_database $DB
		 * @param string $username
		 * @return integer
		 * @throws \moodle_exception
		 */
    public static function stag_get_user_id($username) {
        global $DB;

        $params = self::validate_parameters(self::stag_get_user_id_parameters(),
                array('username' => $username));

        $context = get_context_instance(CONTEXT_SYSTEM);
        self::validate_context($context);
        require_capability('moodle/site:viewuseridentity', $context);

        $user = $DB->get_record('user', array('username'=>$params['username']));
        if ($user == null){
            throw new \moodle_exception('nosuchuser', 'local_stag', '', $params['username']);
				}

        return $user->id;
    }

		/**
		 * @return \external_value
		 */
    public static function stag_get_user_id_returns() {
        return new \external_value(PARAM_INT, "User's identifier");
    }

    /**
     * Unassign role
		 * @return \external_function_parameters
     */
    public static function stag_unassign_role_parameters() {
        return new \external_function_parameters(array(
            'unassign' => new \external_single_structure(array(
                'roleid'    => new \external_value(PARAM_INT, "Role identifier"),
                'course_id' => new \external_value(PARAM_INT, "Course identifier"),
                'user_id'   => new \external_value(PARAM_INT, "User identifier"),
            ))));
    }

		/**
		 * Unassign role
		 * @param array $unassign
		 * @return void
		 */
    function stag_unassign_role($unassign) {
        $params = self::validate_parameters(self::stag_unassign_role_parameters(),
                       array('unassign' => $unassign));
        $roleid = $params['unassign']['roleid'];
        $course_id = $params['unassign']['course_id'];
        $user_id = $params['unassign']['user_id'];

        $contextid = \context_course::instance($course_id);
        $unassignments = array(array(
            'roleid'=>$roleid,
            'userid'=>$user_id,
            'contextid'=>$contextid->__get('id')));

        return \core_role_external::unassign_roles($unassignments);
    }

		/**
		 * @return null
		 */
    public static function stag_unassign_role_returns() {
        return \core_role_external::unassign_roles_returns();
    }

    /*
     * auxiliary routines follow
     */

		/**
		 * @global \moodle_database $DB
		 * @return \stdClass
		 */
    private static function find_course_template() {
        global $DB;

        $cat = $DB->get_record("course_categories", array('name' => self::CATEGORY_TEMPLATE_NAME));
        if (!$cat){
            return null;
				}

        /* XXX make sure there's just a single course in this category */
        $template = $DB->get_record("course", array('category' => $cat->id));

        return $template;
    }

		/**
		 * vraci id takoveho kurzu, ktery ma byt predkem prave vytvareneho
		 * @global moodle_database $DB
		 * @param array $stag_id array(department,subject,year)
		 * @return integer course id
		 */
    private static function find_last_run($stag_id) {
			global $DB;

			$department=$stag_id['department'];
			$subject=$stag_id['subject'];
			$year=(int)$stag_id['year'];

			$attributes=array('department','subject','year','ver','template');
			$attributes_join=array();
			$columns=array();
			foreach($attributes as $attr){
				$attributes_join[]="LEFT JOIN {course_attributes} `".$attr."` ON `".$attr."`.course_id=`course`.`id` AND `".$attr."`.`name`='".$attr."'";
				$columns[]="`".$attr."`.`value` `".$attr."`";
			}
			$query="
				SELECT
					`course`.id,".implode(',',$columns)."
				FROM {course} course
				".implode("\r\n\t\t\t", $attributes_join)."
				WHERE
					`department`.`value`='".$department."' AND
					`subject`.`value`='".$subject."' AND
					`year`.`value`<='".$year."'
				ORDER BY `year`.`value` DESC, `template`.`value` DESC, `ver`.`value` DESC
				LIMIT 1
				;";

			$id=$DB->get_record_sql($query)->id;

			return ($id ? $id : null);

//        $courses = $DB->get_records('course');
//        $last_year = 0;
//        foreach ($courses as $course) {
//            $dep = attribute_manager::get_attribute($course->id, 'department');
//            if ($dep == NULL || $dep != $stag_id['department'])
//                continue;
//
//            $subject = attribute_manager::get_attribute($course->id, 'subject');
//            if ($subject == NULL || $subject != $stag_id['subject'])
//                continue;
//
//            $year = attribute_manager::get_attribute($course->id, 'year');
//            if ($year == NULL || (int)$year > (int)$stag_id['year'])
//                continue;
//
//            if ($last_year < (int)$year) {
//                $last_year = (int)$year;
//                $id = $course->id;
//            }
//        }
//
//        return ($last_year == 0) ? NULL : $id;
    }

		/**
		 * @param integer $course_id
		 * @return void
		 */
    private static function reset_course($course_id) {
        $data = new \stdClass();
        $data->id = $course_id;

        foreach (self::$course_reset_flags as $flag){
            $data->$flag = 1;
				}

        $roles = get_assignable_roles(get_context_instance(CONTEXT_COURSE, $course_id));
        $data->unenrol_users = $roles;

        reset_course_userdata($data);
    }

	/**
	 * vraci unikatni shortname
	 * nejdriv zkusi jestli projde to co je zadano, pak zkusi samotne shortname a pak inkremenuje priponu
	 * to aby slo zadat treba neco jako moje_vlasni_shortname_C
	 * @global moodle_database $DB
	 * @param string $shortname
	 * @param integer $version [optional]
	 * @return string
	 */
	public static function get_unique_shortname($shortname, $version=null, $raw_shortname=null){
		global $DB;
		$new_shortname=false;
		if($shortname){
			if(!$raw_shortname){
				$raw_shortname_arr=explode('_',$shortname);
				$raw_shortname=trim($raw_shortname_arr[0]).(!empty($raw_shortname_arr[1]) ? '_'.trim($raw_shortname_arr[1]) : false).(!empty($raw_shortname_arr[2]) ? '_'.trim($raw_shortname_arr[2]) : false);
			}
			$new_shortname=($version===null ? $shortname : ($raw_shortname.($version ? '_'.$version : false)));
			$result = $DB->get_record_sql("SELECT shortname FROM {course} WHERE shortname=?",array($new_shortname));
			if($result){
				$version=($version===null ? 0 : ($version ? $version+1 : 2));
				$new_shortname=local_stag_course_external::get_unique_shortname($shortname, $version, $raw_shortname);
			}
		}
		else{
			throw new Exception('Shortname must be set!');
		}
		return $new_shortname;
	}
};
